from nmigen import *
from nmigen.cli import main

from functools import reduce

class ExactMatch(Elaboratable):
    def __init__ (self, width, needle):
        self.flush     = Signal(1) # reset state of element

        self.input_bus = Signal(width) # character ingestion port
        self.valid_in  = Signal(1)     # if high, we're ingesting a character from text

        self.match_out  = Signal(1)    # current character matches our predicate

        self.valid_out = Signal(1)     # we're outputting a character from the text
        self.output_bus = Signal(width) # output character

        self.needle = needle

    def elaborate(self, platform):
        m = Module()
        matching = Signal(1)
        valid    = Signal(1)

        m.domain.comb += matching.eq(self.input_bus == self.needle)
        m.domain.comb += valid.eq(self.valid_in)

        m.domain.sync += self.valid_out.eq(valid)
        m.domain.sync += self.output_bus.eq(self.input_bus)
        m.domain.sync += self.match_out.eq(matching & self.valid_in)
        return m



class StringParser(Elaboratable):
    def __init__ (self, needle):

        self.master_match = Signal(1)
        self.needle = needle

    def elaborate(self, platform):
        m = Module()
        curr_idx = Signal(8)

        haystack = Parsebuffer(8,256)
        m.submodules += haystack
        m.domain.sync += curr_idx.eq(curr_idx + 1)
        m.domain.comb += haystack.addr_r.eq(curr_idx)


        eldeen = self.needle[::-1]
        print(eldeen)

        reg_elements = []
        match_signals = []

        for char in eldeen:
            egg = ExactMatch(8, char) #ord(char))
            reg_elements.append(egg)
            m.submodules += egg
            match_signals.append(egg.match_out)

        # wire the chain
        for producer, consumer in zip(reg_elements, reg_elements[1:]):
            m.d.comb += consumer.input_bus.eq(producer.output_bus)
            m.d.comb += consumer.valid_in.eq(producer.valid_out)



        print(len(match_signals))
        # and its ends

        m.d.comb += self.master_match.eq(reduce(lambda x, y: x & y, match_signals))

        m.d.comb += reg_elements[0].input_bus.eq(haystack.data_r)
        m.d.comb += reg_elements[0].valid_in.eq(1)








        return m



class Parsebuffer(Elaboratable):
    def __init__(self, wi, le):
        self.mem = Memory(width=8, depth=256, init=[0x1,0x2,0x3,0x4,0x5, 0x42,0x43,0x44, 0xfe,0xed,0xfa,0xce,0x42,0x00,0x69,0x66,0x6])

        self.data_r = Signal(8)
        self.data_w = Signal(8)

        self.addr_r = Signal(8)
        self.addr_w = Signal(8)

        self.wr_en  = Signal()



    def elaborate(self, platform):
        m = Module()

        m.submodules.parser_port = rdport = self.mem.read_port()
        m.submodules.feeder_port = wport = self.mem.write_port()

        m.domain.comb += [
            rdport.addr.eq(self.addr_r),
            self.data_r.eq(rdport.data), # opposite order, i think because of who's driving who

            wport.en.eq(self.wr_en),
            wport.addr.eq(self.addr_w),
            wport.data.eq(self.data_w)
            ]
        return m


if __name__ == "__main__":
    counter = StringParser([0x42,0x43,0x44])

    main(counter)
