import collections
import functools
import random
from enum import *

#import NetworkX as nx


class TreeNode:
    def __init__(self, language_element, subnodes):
        self.language_element = language_element
        self.subnodes = subnodes

    def __str__(self):
        return str(self.language_element) + " " + str(self.subnodes)


def walk_the_tree(tree, level = 1):
    if tree == None:
        return

    print("    " * (level -1) + "|---" + str(tree.language_element) + " with " + str(len(tree.subnodes)) + " subnodes")
    for subnode in tree.subnodes:
        walk_the_tree(subnode, level + 1)



CFG = collections.namedtuple("CFG", "nonterminals terminals rules start")
# We define a context-free grammar as:
# 1. A list of nonterminal symbols
# 2. A list of terminal symbols
# 3. A list of production rules
# 4. A start symbol.

# Each symbol is defined by a positive integer.


# For the ETF expression grammar:

# Terminal symbols are:
# INTEGER, ADDOP, MULTOP, OPENPAREN, CLOSEPAREN

# Nonterminal symbols are:
# EXPRESSION, TERM, FACTOR

# The start symbol is EXPRESSION

# The rules are:

# EXPRESSION -> EXPRESSION ADDOP TERM
# EXPRESSION -> TERM
# TERM -> TERM MULTOP FACTOR
# TERM -> FACTOR
# FACTOR -> OPENPAREN EXPRESSION CLOSEPAREN
# FACTOR -> INTEGER

#symbols = Enum("Symbols", "EXPRESSION TERM FACTOR INTEGER ADDOP MULTOP OPENPAREN CLOSEPAREN")

nonterminals = Enum("OldNonterminals", "EXPRESSION TERM FACTOR")

terminals = Enum("Terminals", "INTEGER ADDOP MULTOP OPENPAREN CLOSEPAREN")

list_of_nonterminals = [e for e in nonterminals]
list_of_terminals    = [e for e in terminals]

start = nonterminals.EXPRESSION

# The rules are a list of tuples, each of which represents a rule, as follows:

# (Nonterminal symbol, [ordered list of symbols (terminal or nonterminal) that are reduced to aforementioned nonterminal])

rules = [
         (nonterminals.EXPRESSION, [nonterminals.EXPRESSION, terminals.ADDOP,      nonterminals.TERM      ]),
         (nonterminals.EXPRESSION, [nonterminals.TERM                                              ]),
         (nonterminals.TERM,       [nonterminals.TERM,       terminals.MULTOP,     nonterminals.FACTOR    ]),
         (nonterminals.TERM,       [nonterminals.FACTOR                                            ]),
         (nonterminals.FACTOR,     [terminals.OPENPAREN,  nonterminals.EXPRESSION, terminals.CLOSEPAREN]),
         (nonterminals.FACTOR,     [terminals.INTEGER                                           ])
         ]



# We will first implement LR(0), then LR(1), then LALR(1), then IELR(1)

# Here is the LR(0) table generator.

# First we augment the grammar to start with a special start symbol:


def add_starting_symbol(nonterminals, terminals, rules, start):
    if ("START" in [e for e in nonterminals]):
        return None
    new_nonterminal_list = [m.name for m in nonterminals]+ ['START']
    nonterminals_new = Enum("newNonterminals", new_nonterminal_list)

    the_new_rules = []

    for rule in rules:
        rewritten_RHS = []

        assert(isinstance(rule[0], nonterminals))

        new_lhs = nonterminals_new(rule[0].value)


        for langitem in rule[1]:
            if isinstance(langitem, nonterminals):
                newlangitem = nonterminals_new(langitem.value)
            else:
                newlangitem = langitem
            rewritten_RHS.append(newlangitem)

        the_new_rules.append([new_lhs, rewritten_RHS])

    assert(isinstance(start, nonterminals))

    not_really_start_symbol = nonterminals_new(start.value)
    final_rule = [nonterminals_new.START, [not_really_start_symbol]]
    the_new_rules.append(final_rule)
    return (nonterminals_new, the_new_rules)




(nnt, newrules) = add_starting_symbol(nonterminals, terminals, rules,start)


print(newrules)








