from nmigen import *
from nmigen.cli import main

from nmigen.back.pysim import Simulator, Delay

from functools import reduce


from typing import List, Dict, Tuple, Optional



class DummyPlug(Elaboratable):
    def __init__(self):
        self.source_memory_addr = Signal(8)
        self.source_memory_data = Signal(8)
        self.ctr = Signal(8)
        self.wdata = Signal(8)


    def elaborate(self, platform):
        m = Module()

        ctr = Signal(8)

        mem = Memory(width=8, depth=16)
        self.introspector = mem
        m.submodules.output_memory = wport = mem.write_port()

        m.d.comb += (wport.addr).eq(ctr)
        m.d.comb += (wport.en).eq(1)
        m.d.comb += (wport.data).eq(self.source_memory_data + 3)
        m.d.sync += ctr.eq(ctr + 1)
        m.d.comb += self.source_memory_addr.eq(ctr)

        m.d.comb += self.ctr.eq(ctr)
        m.d.comb += self.wdata.eq(wport.data)

        return m

if __name__ == '__main__':
    m = Module()


    m.submodules.baka = nine = DummyPlug()

    for master_idx in range(4):
        print("MASTER IDX IS", master_idx)
        trace = []
        def process():
            for _ in range(16):
                yield
                array = []
                for idx in range(16):
                    print(idx)
                    x = yield nine.introspector[idx]
                    array.append(x)
                trace.append(array)

        mem = {
        0x00: 0x40,
        0x01: 0x41,
        0x02: 0x42,
        0x03: 0x43
        }
        with m.Switch(nine.source_memory_addr):
            for addr,data in mem.items():
                with m.Case(addr):
                    m.d.comb += nine.source_memory_data.eq(data + master_idx)
            with m.Default():
                    m.d.comb += nine.source_memory_data.eq(0xfe)


        sim = Simulator(m)
        sim.add_clock(1e-9)
        sim.add_sync_process(process)
        with sim.write_vcd("test.vcd", "test.gtkw"):
            sim.run()

        print(trace)







